from .DataObject import DataObject
from .MetadataObject import MetadataObjectCollection
from .Polymorphic import Polymorphic
import re


class Entity(Polymorphic, DataObject):
    
    def __init__(self, schema=[]):
        super().__init__(schema)

    invalidEntityNameRegex = re.compile("\\.|\"")

    def validate(self):
        super().validate()
        if self.invalidEntityNameRegex.search(self.name):
            raise ValueError("%s.name cannot contain dot or quotation mark." % (self.__class__.__name__, ))


class EntityCollection(MetadataObjectCollection):
    itemType = Entity

    @classmethod
    def fromJson(cls, entities):
        result = EntityCollection()
        for entity in entities:
            result.append(Entity.fromJson(entity))
        return result


from .Base import Base
from .SchemaEntry import SchemaEntry
from .ObjectCollection import ObjectCollection
from .utils import String
from .utils import Uri


class Reference(Base):
    def __init__(self, schema=[]):
        self.schema = schema + (
            SchemaEntry("id", String),
            SchemaEntry("location", Uri)
        )
        super().__init__(self.schema)

class ReferenceCollection(ObjectCollection):
    itemType = Reference


class PolymorphicMeta(type):
    classes = {}
    
    def __new__(cls, name, bases, attrs):
        cls = type.__new__(cls, name, bases, attrs)
        cls.classes[cls] = {cls.__name__ : cls}
        cls.__appendBases(bases, cls)
        return cls
    
    @staticmethod
    def __appendBases(bases, cls):
        for base in bases:
            basemap = cls.classes.get(base, None)
            if basemap is not None:
                basemap[cls.__name__] = cls
                cls.__appendBases(base.__bases__, cls)

class Polymorphic(metaclass=PolymorphicMeta):
    @classmethod
    def fromJson(cls, value, _=""):
        actualClass = PolymorphicMeta.classes[cls][value["$type"]]
        schema = actualClass().getSchema()
        return super(Polymorphic, actualClass).fromJson(value, schema)




from .Writer import Writer
from azure.storage.blob import BlockBlobService
import json
import pandas as pd
import numpy as np


class ADLSWriter(Writer):
    """
        Write dataframe/json to specified blob storage location
    """
    def __init__(self, account_name, account_key, container_name, storage_name, dataflow_name):
        self.account_name = account_name
        self.account_key = account_key
        self.container_name = container_name
        self.storage_name = storage_name
        self.dataflow_name = dataflow_name

    def create_snapshot(self, location, snapshot_dir_name):
        """
            Create a snapshot of the file current file in the passed directory name
            location: dir1/dir2/dir3/filename.extension or "model.json"
            snashot_dir_name: Snapshot directory name
        """
        if location.strip() == "model.json":
            location = self.dataflow_name + "/" + location

        exists = block_blob_service.exists(self.container_name, location)
        if not exists:
            return False

        t = datetime.datetime.now().strftime('%d-%M-%Y-%H-%M-%S')
        file_path, filename = '/'.join(location.split('/')[:-1]), location.split('/')[-1]
        
        # copy file into snapshot folder
        old_blob_url = block_blob_service.make_blob_url(self.container_name, location)
        block_blob_service.copy_blob(self.container_name, file_path + '/' + snapshot_folder_name + '/' + filename + '@snapshot' + t, old_blob_url)
        new_blob_url = block_blob_service.make_blob_url(container_name, file_path + '/' + snapshot_folder_name + '/' + filename + '@snapshot' + t)
        
        # Delete old file
        block_blob_service.delete_blob(container_name, filename)
        return True
    
    def write_df(self, blob_location, dataframe, number_of_partition=5):
        """
            Write dataframe to specified blob storage location
        """
        block_blob_service = BlockBlobService(account_name=self.account_name, account_key=self.account_key)
        dfs = np.array_split(dataframe, number_of_partition)
        result = list()

        entity_name = blob_location.split('/')[0]
        blob_location = blob_location + "/" + entity_name
        for i in range(len(dfs)):
            dataframe = dfs[i].to_csv(index=False, header=False)
            filename = blob_location + str(i) + ".csv"
            block_blob_service.create_blob_from_text(self.container_name + "/" + self.dataflow_name, 
                                                     filename, dataframe)
            blob_url = 'https://' + self.storage_name + '.dfs.core.windows.net/' + self.container_name + '/' + filename
            result.append((filename, blob_url))
        return result

    def write_json(self, blob_location, json_dict):
        """
        write json to specified blob storage location
        """
        json_dict = json.dumps(json_dict)
        block_blob_service = BlockBlobService(
            account_name=self.account_name, account_key=self.account_key)
        block_blob_service.create_blob_from_text(self.container_name, self.dataflow_name+"/"+ blob_location, json_dict)
        blob_url = 'https://'+self.storage_name+'.dfs.core.windows.net/'+self.container_name+'/'+blob_location
        return blob_url



String = str
DateTimeOffset = str
Uri = str
import datetime
import pytz

def getattrIgnoreCase(obj, attr, default=None):
    for i in dir(obj):
        if i.lower() == attr.lower():
            return getattr(obj, attr, default)
    return default


dtype_converter = {
    'int': 'int64',
    'bigint': 'int64',
    'long': 'int64',
    'float':'decimal',
    'double': 'double',
    'decimal.Decimal': 'decimal',
    'string': 'string',
    'bool': 'boolean',
    'datetime': 'dateTime',
    'timestamp': 'dateTime'
}


def to_utc_timestamp(d, format, tz=None):  
    if d is None or d == '':
        return ''  
    if not isinstance(d, datetime.datetime):
        d = datetime.datetime.strptime(d, format)
    if tz:
        tz = pytz.timezone(tz)
        utc_time = tz.normalize(tz.localize(d)).astimezone(pytz.utc)
    else:
        utc_time = d.astimezone(pytz.utc)
    return str(int(utc_time.timestamp()))


def from_utc_timestamp(d, format, tz=None, offset_hour=False):
    if d is None or d == '' or d != d:
        return ''  
    d = str(d)
    if not offset_hour:
        tz = pytz.timezone(tz)
        utc_time = datetime.datetime.utcfromtimestamp(int(d.split(".")[0]))
        converted_time = tz.normalize(pytz.utc.localize(utc_time)).astimezone(tz)
    else:
        offset_hour = tz
        hour, minutes = map(int, offset_hour.split(":"))
        if hour < 0:
            days = -1
            total_second = 86400 - (((hour - 2*hour) * 60 + minutes) * 60)
        else:
            days = 0
            total_second = (hour * 60 + minutes) * 60
        converted_time = datetime.datetime.fromtimestamp(int(d.split(".")[0]), 
                            datetime.timezone(datetime.timedelta(days=days, seconds=total_second)))
    return converted_time.strftime(format)


from .Base import Base
from .SchemaEntry import SchemaEntry
from .utils import String

class AttributeReference(Base):
    def __init__(self, schema=[]):
        self.schema = schema + [
            SchemaEntry("entityName", String),
            SchemaEntry("attributeName", String)
        ]
        super().__init__(self.schema)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.entityName == other.entityName and self.attributeName == other.attributeName
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def validate(self):
        super().validate()
        className = self.__class__.__name__
        if not self.entityName:
            raise ValueError("%s.entityName is not set" % (className, ))
        if not self.attributeName:
            raise ValueError("%s.attributeName is not set" % (className, ))

    def getSchema(self):
        return super().getSchema()

    @classmethod
    def fromJson(cls, value):
        schema = AttributeReference().getSchema()
        return super().fromJson(value, schema)


from .utils import getattrIgnoreCase
from collections import OrderedDict


class Base(object):

    __ctors = {}

    def __init__(self, schema):
        self.schema = schema
        for entry in schema:
            setattr(self, entry.name, entry.defaultValue)

    def getSchema(self):
        return self.schema

    def toJson(self):
        result = OrderedDict()
        for entry in self.schema:
            element = getattrIgnoreCase(self, entry.name, result)
            if element != result and entry.shouldSerialize(element):
                result[entry.name] = getattr(element, "toJson", lambda: element)()
        result.update(getattrIgnoreCase(self, "customProperties", {}))
        return result

    def validate(self):
        tmp = object()
        className = self.__class__.__name__
        for entry in self.schema:
            element = getattrIgnoreCase(self, entry.name, tmp)
            if element != tmp and element is not None:
                if not isinstance(element, entry.cls):
                    raise TypeError("%s.%s must be of type %s" % (className, entry.name, entry.cls))
                getattr(element, "validate", lambda: None)()

    @classmethod
    def fromJson(cls, value, schema):
        result = cls()
        for entry in schema:
            element = value.pop(entry.name, result)
            if element != result:
                setattr(result, entry.name, cls.__getCtor(entry.cls)(element))
        result.customProperties = value
        return result
    
    @classmethod
    def __getCtor(cls, type):
        ctor = cls.__ctors.get(type, None)
        if not ctor:
            ctor = getattr(type, "fromJson", type)
            cls.__ctors[type] = ctor
        return ctor




from .Model import Model
from .Writer import *
from .Reader import *
from .Annotation import *
from .Attribute import *
from .AttributeReference import *
from .CdmDataType import *
from .CsvFormatSettings import *
from .DataObject import *
from .Entity import *
from .LocalEntity import *
from .Partition import *
from .Reference import *
from .ReferenceEntity import *
from .Relationship import *
from .SingleKeyRelationship import *
from .SchemaEntry import *
from .utils import *
from .ADLSWriter import *
from .ADLSReader import *
from .ADLSPandasReader import *